#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   mktoc.base
   ~~~~~~~~~~

   Standard defines used by all modules.

   All modules in this package import this file into its root namespace.
"""

__all__        = ['__author__', '__copyright__', '__email__', '__license__',
                  'VERSION', 'MkTocError' ,'FileNotFoundError',
                  'TooManyFilesMatchError', 'ParseError', 'UnderflowError',
                  'EmptyCueData' ]

#: Project author string.
__author__     = 'Patrick C. McGinty'
#: Project copyright string.
__copyright__  = 'Copyright (c) 2011'
#: Project e-mail address.
__email__      = 'mktoc[@]tuxcoder[dot]com'
#: Project license string.
__license__    = 'BSD'
#: Project version number string.
VERSION        = '1.2.2'

class MkTocError(Exception):
   """A base exception class for all mktoc exceptions classes."""
   pass

class FileNotFoundError(MkTocError):
   """Exception class used whenever a file can not be located and is
   required by the system."""
   pass

class TooManyFilesMatchError(MkTocError):
   """Exception class used whenever multiple files are found that match the
   source file required by the system."""
   def __init__(self,src_file,found_files):
      self.src_file = src_file
      self.found_files = found_files

class ParseError(MkTocError):
   """Exception class indicates that a CUE file could not be parsed do to
   unknown data."""
   pass

class UnderflowError(MkTocError):
   """Exception class used by the disc module to indicate that a track
   time subtraction has caused and underflow."""
   pass

class EmptyCueData(MkTocError):
   """Exception class indicates that the input CUE file or STDIN data was
   empty.  The mktoc application interprets this as a normal condition and
   does not raise an error to the user other than returning a non-zero
   exit code."""
   pass




#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   mktoc.disc
   ~~~~~~~~~~

   A set of classes for representation of audio CD information.

   The following are a list of the classes provided in this module:

   * :class:`Disc`
   * :class:`Track`
   * :class:`TrackIndex`
"""

import os
import re
import wave
import logging
import itertools as itr

from mktoc.base import *

__all__ = [ 'Disc', 'Track', 'TrackIndex' ]

log = logging.getLogger('mktoc.disc')


class Disc( object ):
   """
   Stores audio disc metadata values such as album title, performer, genre.
   """
   #: Disc mode string for single session
   MODE_SINGLE_SESSION  = 'CD_DA'
   #: Disc mode string for multi sessions
   MODE_MULTI_SESSION   = 'CD_ROM_XA'
   #: String representing the Catalog Id of a disc.
   catalog   = None
   #: String representing the release year of a disc.
   date      = None
   #: String representing the DiscID value of a disc. This value is assumed to
   #: be correct and not verified by any internal logic.
   discid    = None
   #: String representing the genre of a disc
   genre     = None
   #: String representing the performer or artist of a disc.
   performer = None
   #: String representing the album title of a disc.
   title     = None

   def __init__(self):
      self.is_multisession = False

   def __str__(self):
      """Return a string of TOC formatted disc information."""
      out = ['%s' % self._mode]
      if self.catalog:   out += ['CATALOG "%s"' % self.catalog]
      out += ['CD_TEXT { LANGUAGE_MAP { 0:EN }\n\tLANGUAGE 0 {']
      if self.title:     out += ['\t\tTITLE "%s"' % self.title]
      if self.performer: out += ['\t\tPERFORMER "%s"' % self.performer]
      if self.discid:    out += ['\t\tDISC_ID "%s"' % self.discid]
      out += ['}}']
      return '\n'.join(out)

   def set_field(self, name, value):
      """
      Set a disc field value as a class attributes.

      This method provides additional formating to any data fields. For
      example, removing quoting and checking the field name.

      :param name: Field name to set
      :type  name: str

      :param value: Value of field
      :type  value: str

      :return: :data:`True` if field is written to the class data, or
               :data:`False`
      """
      name = name.lower()
      if hasattr(self,name):
         setattr(self,name,value.strip('"'))
         return True
      return False

   @property
   def is_multisession(self):
      """
      :return: :data:`True` if disc is defined as multi-sesssion.
      """
      return self._mode == self.MODE_MULTI_SESSION

   @is_multisession.setter
   def is_multisession(self,val):
      """
      Change the mode of a disc to either 'single-session' or 'multi-session'.

      :param val: :data:`True` if disc is multi-session, :data:`False`
                  otherwise.
      :type  val: bool
      """
      if val:  self._mode = self.MODE_MULTI_SESSION
      else:    self._mode = self.MODE_SINGLE_SESSION


class Track( object ):
   """
   Holds track metadata values such as title and performer. Each :class:`Track`
   object contains a list of :class:`TrackIndex`\s that specifies the
   audio data associated with the track.
   """

   #: :data:`True` or :data:`False`, indicates *Digital Copy Protection*
   #: flag on  the track.
   dcp          = False
   #: :data:`True` or :data:`False`, indicates *Four Channel Audio* flag
   #: on the track.
   four_ch      = False
   #: list of :class:`TrackIndex` objects. Every track has at
   #: least one :class:`TrackIndex` and possibly more. The
   #: :class:`TrackIndex` defines a length of audio data or property in
   #: the track. The fist :class:`TrackIndex` can be pre-gap data. Only
   #: one audio file can be associated with a :class:`TrackIndex`, so if a
   #: track is composed of multiple audio files, there will be an >=
   #: number of :class:`TrackIndex`\s.
   indexes      = None
   #: :data:`True` or :data:`False`, indicates if a track is binary data
   #: and not audio. Data tracks will not produce any text when printed.
   is_data      = False
   #: String representing ISRC value of the track.
   isrc         = None
   #: Integer initialized to the value of the track number.
   num          = None
   #: String representing the track artist.
   performer    = None
   #: :data:`True` or :data:`False`, indicates *Pre-Emphasis* flag on the
   #: track.
   pre          = False
   #: :class:`_TrackTime` value that indicates the pre-gap value of the
   #: current track. The pre-gap is a time length at the beginning of a
   #: track that will cause a CD player to count up from a negative time
   #: value before changing the track index number.  The starting pre-gap
   #: value of a track is essentially the final audio at the end of the
   #: previous track.  However, there is more than one way to designate
   #: the pregap in a track, therefore this variable is only used if the
   #: first :class:`TrackIndex` in the track contains more than just the
   #: pre-gap audio.
   pregap       = None
   #: String representing the title of the track.
   title        = None

   def __init__(self,num,is_data=False):
      """
      :param num: Track index in the audio CD.
      :type num:  int

      :param is_data: :data:`True` if track is data instead of audio.
      :type  is_data: bool
      """
      # create an empty list of :class:`TrackIndex` objects, and
      # assign track number.
      self.indexes   = []    # list of indexes in the track
      self.num       = num
      self.is_data   = is_data

   def __str__(self):
      """Return the TOC formated representation of the :class:`Track`
      object including the :class:`TrackIndex` objects. Data tracks will
      not generate any output."""
      if self.is_data:
         return ''     # do not output to TOC
      out = ['\n//Track %d' % self.num]
      out += ['TRACK AUDIO']
      if self.isrc:       out += ['\tISRC "%s"' % self.isrc]
      if self.dcp:        out += ['\tCOPY']
      if self.four_ch:    out += ['\tFOUR_CHANNEL_AUDIO']
      if self.pre:        out += ['\tPRE_EMPHASIS']
      out += ['\tCD_TEXT { LANGUAGE 0 {']
      if self.title:      out += ['\t\tTITLE "%s"' % self.title]
      if self.performer:  out += ['\t\tPERFORMER "%s"' % self.performer]
      out += ['\t}}']
      if self.pregap:     out += ['\tPREGAP %s' % self.pregap]

      for idx in self.indexes:
         out.append( str(idx) )

      return '\n'.join(out)

   def set_field(self, name, value):
      """
      Set a track field value as a class attributes.

      This method provides additional formating to any data fields. For
      example, removing quoting and checking the field name.

      :param name: Field name to set
      :type  name: str

      :param value: Value of field
      :type  value: str

      :return: :data:`True` if field is written to the class data, or
               :data:`False`
      """
      name = name.lower()
      if hasattr(self,name):
         if isinstance(value,basestring):
            setattr(self,name,value.strip('"'))
         elif isinstance(value,bool):
            setattr(self,name,value)
         return True
      return False


class TrackIndex(object):
   """
   Represent an *index* of an audio CD track.

   Specifically, information about a location, length and type of audio data.
   :class:`Track` objects can have one or more :class:`TrackIndex`\s to
   represent the audio data belonging to the track.

   .. rubric:: Constants

   .. data:: PREAUDIO

      Indicates a :class:`TrackIndex` that is pre-gap audio data only.

   .. data:: AUDIO

      Indicates a :class:`TrackIndex` of standard audio data or both pre-gap
      and audio.

   .. data:: INDEX

      Indicates a :class:`TrackIndex` after the start of a time stamp of a
      previous :const:`AUDIO` :class:`TrackIndex` object. There is no audio
      data associated with *INDEX* :class:`TrackIndex`\s.

   .. data:: START

      Same as :const:`INDEX`, but :class:`TrackIndex` preceding it is was
      the pre-gap audio of the same :class:`Track`.

   .. rubric::  Attributes

   .. attribute:: file_

      String representing a WAV file's path and name. This is used to read the
      audio data of the :class:`TrackIndex`.

   .. attribute:: len_

      Empty string or :class:`_TrackTime` value that specifies the number of
      audio frames associated with the :class:`TrackIndex`. By default, this
      value will equal the total length of the WAV data, but might be truncated
      if the track starts after, or ends before the WAV data.

   .. attribute:: num

      Integer specifying the location of the :class:`TrackIndex` in the track.
      The first index *num* is always 0.

   .. attribute:: time

      :class:`_TrackTime` value that specifies the starting time index of the
      :class:`TrackIndex` object relative to the start of the audio data.
      Usually this value is ``0``.
   """

   #: Enum of valid :class:`TrackIndex` types.
   PREAUDIO, AUDIO, INDEX, START, DATA = range(5)

   #: Integer set to :const:`PREAUDIO` or :const:`AUDIO` or :const:`INDEX` or
   #: :const:`START`. Indicate the mode of :class:`TrackIndex` object.
   cmd = AUDIO

   def __init__(self, num, time, file_, len_=None):
      """
      If possible the sample count of the :class:`TrackIndex` is calculated by
      reading the WAV audio data.

      :param num:    Index number position in the :class:`Track`, starting
                     at 0.
      :type num:     int

      :param time:   The indexes starting offset in the audio data file.
      :type  time:   :class:`_TrackTime`

      :param file_:  String representing the path location of a WAV file
                     associated with this index.
      :type  file_:  str

      :param len_:   Track length in format supported by :class:`_TrackTime`.
      :type  len_:   str, tuple, int (see :class:`_TrackTime`)
      """
      self.file_  = file_
      self.num    = int(num)
      self.time   = _TrackTime(time)
      if len_:
         self.len_ = _TrackTime(len_)
      else:
         # set length to maximum possible for now (total - start)
         file_len = self._file_len(self.file_)
         if file_len: self.len_ = file_len - self.time
         else:        self.len_ = ''
      log.debug( 'creating index %s' % repr(self) )

   def __repr__(self):
      """Return a string used for debug logging."""
      return "'%s, %s'" % (self.file_, self.time)

   def __str__(self):
      """Return the TOC formated string representation of the
      :class:`TrackIndex` object."""
      out = []
      if self.cmd == self.DATA:
         return ''     # do not output to TOC
      if self.cmd in [self.AUDIO, self.PREAUDIO]:
         out += ['\tAUDIOFILE "%(file_)s" %(time)s %(len_)s' % self.__dict__]
      elif self.cmd == self.INDEX:
         out += ['\tINDEX %(time)s' % self.__dict__]
      elif self.cmd == self.START:
         out += ['\tSTART %(len_)s' % self.__dict__]
      else: raise Exception
      # add start command for pregap audio
      if self.cmd == self.PREAUDIO:
         out += ['\tSTART']
      return '\n'.join(out)

   def _file_len(self,file_):
      """Returns the number of audio samples in the WAV file, *file_*.
      Called during __init__. If *file_* can not be opened, :data:`None`
      is returned.

      :param file_:  a file name string relative to the cwd referencing
                     a WAV file.
      :type file_:   str

      :rtype:        :class:`_TrackTime` of audio samples or
                     :data:`None`"""

      if not (file_ and os.path.exists(file_)):
         return None
      w = wave.open(file_)
      frames = w.getnframes() / (w.getframerate()/75)
      w.close()
      return _TrackTime(frames)


class _TrackTime(object):
   """
   Container class to represent the sample count or position in audio data.
   Allows mathematical operations to be easily performed on time positions.
   """
   #: Defines the number of audio *Frames Per Second*
   _FPS = 75
   #: Defines the number of *Seconds Per Minute*
   _SPM = 60
   #: Defines the number of audio *Frames Per Minute*
   _FPM = _FPS * _SPM

   #: :class:`tuple` that stores the minutes, seconds, and frames values. The
   #: combination of these values can be used to calculate the total frame
   #: count.
   _time = None

   def __init__(self, arg=None):
      """Initializes the :class:`_TrackTime` object, normalizing the input
      data.

      :param arg: Variable representation of the value of the
                  :class:`_TrackTime` object. The allowed formats are:
                     a. String in the format *MM:SS:FF*
                     b. Tuple in the format (M,S,F)
                     c. Integer of the total frame length
                     d. :data:`None`, object is initialized to 0 length
      :type arg: str, :class:`tuple`, int, :data:`None`"""
      if isinstance(arg,basestring):
         # extract time from string
         val = [int(x) for x in arg.split(':')]
         self._time = tuple(val)
      elif isinstance(arg,tuple):
         # assume arg is correct format
         self._time = arg
      elif isinstance(arg,(int,long)):
         # convert frame count to min,sec,frames
         min_,fr = divmod(arg, self._FPM)
         sec,fr = divmod(fr, self._FPS)
         self._time = (min_,sec,fr)
      else:
         # set time to '0:0:0' (zero)
         self._time = tuple([0]*3)

   def __repr__(self):
      """Return string value of the :class:`_TrackTime` in format
      *MM:SS:FF*."""
      return '%02d:%02d:%02d' % self._time

   def __ne__(self, other):
      """Return :data:`True` if objects are NOT equal."""
      return self._time != other._time

   def __eq__(self, other):
      """Return :data:`True` if objects are equal."""
      return self._time == other._time

   def __sub__(self, other):
      """Return result of *self* - *other*."""
      mn,sc,fr = map(lambda x,y: x-y, self._time, other._time)
      if fr<0: sc-=1; fr+=self._FPS
      if sc<0: mn-=1; sc+=self._SPM
      if mn<0: raise UnderflowError, \
         'Track time calculation resulted in a negative value'
      return _TrackTime((mn,sc,fr))

   @property
   def frames(self):
      """Convert min,sec,frame to total frames."""
      return sum( [x*y for x,y in
                     zip(self._time,[self._FPM,self._FPS,1]) ])




#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   .. Mktoc // (c) 2011, Patrick C. McGinty
   .. mktoc[@]tuxcoder[dot]com

   Mktoc simplifies the steps needed to create audio CD TOC files for the
   cdrdao CD burning program. For users familiar with ExactAudioCopy or CdrWin,
   TOC files are synonymous with CUE sheets. The primary goal of mktoc is to
   create TOC files using a previously generated CUE sheet.

   Features
   ========

   * Create a TOC file from a list of WAV files.
   * Convert an ExactAudioCopy (EAC) CUE file to the TOC format that
     is usable by cdrdao.
   * Multi-session aware CUE parsing, necessary for making CDDB and
     AccurateRip matchable discs.
   * Non-compliant CUE sheet support.
   * Support for various pregap methods.
   * Can create offset corrected WAV files for true 'bit-for-bit'
     accurate copies.
   * Fuzzy file name logic can correct common file name spelling
     variations.
   * Workaround known TOC file parsing bugs in cdrdao.

   Usage
   =====
   ::

      mktoc [OPTIONS] < CUE_FILE
      mktoc [OPTIONS] [[-f] CUE_FILE] [[-o] TOC_FILE]
      mktoc [OPTIONS] -w WAV_FILES [[-o] TOC_FILE]

   ``CUE_FILE`` must contain a valid CUE format. When ``*_FILE`` is not
   provided, the program will read from ``STDIN``. All output will be sent to
   ``STDOUT``.

   All attempts will be made to preserve any and all information from the input
   file. For any case where the CUE file contains unknown or bad values, the
   user will be notified on ``STDERR``.

   Options
   =======

   --version

      show program's version number and exit

   -h

      show help message and exit

   --help

      show detailed usage instructions and exit

   -a, --allow-missing-wav

      do not abort when WAV file(s) are missing, (experts only). It is possible
      when using this option that a bug in cdrdao will create a CD that ignores
      the pregap definitions in the TOC file.  Only use this option if the CUE
      file does not contain pregaps, or if you do not wish to retain the pregap
      information.

   -c <WAV_OFFSET>, --offset-correction=<WAV_OFFSET>

      correct reader/writer offset by creating WAV file(s) shifted by
      WAV_OFFSET samples (original data is not modified)

   -d, --debug

      enable debugging statements

   -f <CUE_FILE>, --file=<CUE_FILE>

      specify the input CUE file to read

   -m, --multi

      for safety, this option must be set when creating a mulit-session TOC
      file

   -o <TOC_FILE>, --output=<TOC_FILE>

      specify the output TOC file to write

   -t, --use-temp

      write offset corrected WAV files to /tmp directory

   -w, --wave

      write a TOC file using list of WAV files

   -z, --no-multi

      disable multi-session support; program assumes TOC will be written in
      single-session mode

   Examples
   ========

   1. Create a TOC file from a set of WAV files::

         mktoc -w *.wav

   2. Write a TOC file to ``toc_file.toc``, from a set of WAV files::

         mktoc -w *.wav toc_file.toc
         mktoc -w *.wav > toc_file.toc
         mktoc -w *.wav -o toc_file.toc

   3. Create a TOC file from a valid CUE file::

         mktoc cue_file.cue
         mktoc < cue_file.cue
         mktoc -f cue_file.cue

   4. Write a TOC file to ``toc_file.toc``, given an input CUE file::

         mktoc cue_file.cue toc_file.toc
         mktoc < cue_file.cue > toc_file.toc
         mktoc -f cue_file.cue -o toc_file.toc

   5. Tell mktoc to ignore missing WAV file errors. There is a potential that
      the result TOC file will cause cdrdao to loose pregap information during
      the burn process (see above)::

         mktoc -a cue_file.cue

   6. Adjust WAV files for a CD writer offset value. For example, if your CD
      writer has a -30 sample write offset, it can be corrected by offsetting
      the input WAV files by +30 samples. New WAV files will be placed in the
      working directory in a new dir called ``wav+30``::

         mktoc -c 30 < cue_file.cue

   7. Adjust WAV files for a CD writer offset value, but create new files in
      the ``/tmp`` directory::

         mktoc -c 30 -t < cue_file.cue

   Contact
   =======

   E-mail
   ------

   |  mktoc[@]tuxcoder[dot]com

   Info
   ----

   |  http://packages.python.org/mktoc/
   |  https://github.com/cmcginty/mktoc
   |  http://blog.tuxcoder.com
"""

import sys
import mktoc.cmdline

if __name__ == '__main__':
   sys.exit(mktoc.cmdline.main())



#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   mktoc.wav
   ~~~~~~~~~

   Utility classes for search and modifying WAV audio files.

   The following are a list of the classes provided in this module:

   * :class:`WavFileCache`
   * :class:`WavOffsetWriter`
"""

import os
import sys
import re
import wave
import tempfile
import logging
import itertools as itr
import operator as op

from mktoc.base import *

__all__ = ['WavFileCache', 'WavOffsetWriter']

log = logging.getLogger('mktoc.wav')


##############################################################################
class WavFileCache(object):
   """
   Verifies the existence of a WAV file in the local file system.

   The class provides fuzzy logic name matching of cached results in the case
   that the specified file can not be found. The files system is only scanned
   once, and all lookups after the initial test come from the cache. The cache
   size is limited to prevent over aggressive file system access.
   """

   # list of WAV files found in the local file system.
   _data = None

   # base search path location.
   _src_dir = None

   # complied search object that can be used to match file name strings ending
   # with the '.wav' extension.
   _WAV_REGEX = re.compile(r'\.wav$', re.IGNORECASE)

   def __init__(self, _dir=os.curdir):
      """
      Initialize the class instance with the input :attr:`_dir` argument. If no
      argument is supplied it defaults to the current working dir.

      :param _dir:   Base path location to perform the WAV file search.
      :type _dir:    str

      .. Docuemnt private members
      .. automethod:: __call__
      """
      assert(_dir)
      self._src_dir = _dir

   def __call__(self, file_):
      """
      Search the cache for a fuzzy-logic match of the file name in
      :attr:`file_` parameter. This method will always return the exact file
      name if it exists before attempting fuzzy matches.

      :param file_:  File name to search for.
      :type file_:   str
      """
      log.debug("looking for file '%s'",file_)
      tmp_name = file_
      # convert a DOS file path to Linux
      tmp_name = tmp_name.replace('\\','/')
      # base case: file exists, and is has a 'WAV' extension
      if self._WAV_REGEX.search(tmp_name) and os.path.exists(tmp_name):
         log.debug('-> FOUND\n'+'-'*5)
         return file_       # return match
      # case 2: file is locatable in path by stripping directories
      fn = os.path.basename(tmp_name)     # strip leading path
      fn = os.path.splitext(fn)[0]        # strip extension
      fn = fn.strip()                     # strip any whitespace
      log.debug("-> looking for file '%s'", os.sep + fn + '.wav')
      # escape any special characters in the file, and the '$' prevents
      # matching if any extra chars come after the name
      sep = re.escape(os.sep)
      fn_pat = sep + '.*' + re.escape(fn) + '.*'
      fn_pats = [fn_pat]
      # same as pat1, but replace spaces with underscores
      fn_us = fn.replace(' ','_')
      fn_pat = sep + '.*' + re.escape(fn_us) + '.*'
      fn_pats.append( fn_pat )
      # same as pat1, but replace underscores with spaces
      fn_us = fn.replace('_',' ')
      fn_pat = sep + '.*' + re.escape(fn_us) + '.*'
      fn_pats.append( fn_pat )
      file_regex = re.compile( '|'.join(set(fn_pats)), re.IGNORECASE)
      # search all WAV files using pattern 'file_regex'
      matchi = itr.imap( file_regex.search, self._get_cache() )
      # create tuple with input file and search results
      matches = itr.izip( self._get_cache(), matchi )
      matches = filter( op.itemgetter(1), matches )
      if len(matches) == 1:   # success if ONE match is found
         log.debug("--> FOUND '%s'" % matches[0][0])
         return matches[0][0]
      elif len(matches) == 0:
         raise FileNotFoundError, file_ # zero or multiple matches is an error
      else:
         raise TooManyFilesMatchError, (file_, [m[0] for m in matches])

   def _get_cache(self):
      """
      Helper function used to lookup the WAV file cache. The first call to this
      method will cause the creation of the cache.
      """
      if self._data is None:
         self._init_cache()
      return self._data

   def _init_cache(self):
      """
      Create a list of WAV files in the vicinity of the current working dir.
      The list is store in the object member '_data'.
      """
      self._data = []
      fc = 0
      log.debug("Initializing file cache @ '%s'", self._src_dir)
      for root, dirs, files in os.walk(self._src_dir):
         if fc > 1000: break     # only cache first n files
         fc += len(files)
         f_tup = zip( [root]*len(files), files )
         wav_files = [os.path.join(r,f) for r,f in f_tup \
                           if self._WAV_REGEX.search(f)]
         self._data.extend( wav_files )
      self._is_init = True
      log.debug('-> Found %d files:' % len(self._data) )
      map( lambda f: log.debug('--> %s' % f), self._data )


##############################################################################
class WavOffsetWriter(object):
   """
   Shift the audio data in a set of WAV files by a desired postive or negative
   sample offset.

   The module will never modify the input WAV files, and always write to either
   a new directory in the 'cwd' or in the :file:`/tmp` directory. The WAV files
   are treated as a set of data, in that, the direction of shift will cause
   audio sample data to be taken from either a previous or next WAV file. The
   shift in sample data will cause either the first or last WAV file to contain
   'sample count' of NULL samples.
   """

   # number of samples to copy for each cycle. This value affects the memory
   # required by this class and the frequency the progress bar is update.
   _COPY_SIZE = 256*1024

   # sample shift offset value.
   _offset = None

   # reference to a :class:`ProgressBar` instance to provide progress updates.
   _pb = None

   # string of the program name (i.e. mktoc) used when creating directories in
   # /tm.
   _progName = None

   def __init__(self, offset_samples, pb_class, pb_args):
      """
      :param offset_samples:  Sample shift value
      :type offset_samples:   int

      :param pb_class:  outputs status updates to the user. First argument of
                        the class init routine specifies the maximum value of
                        the progress bar and is calulated by this class.
      :type pb_class:   :class:`ProgressBar`

      :param pb_args:   Argument list used to initialize progress bar. However,
                        the first argument of the progress bar init routine is
                        calculated by this class.
      :type pb_args:    list

      .. Document private members
      .. automethod:: __call__
      """
      self._offset  = offset_samples
      self._pb_class = pb_class
      self._pb_args  = pb_args
      self._progName = os.path.basename( sys.argv[0] )

   def __call__(self, files, use_tmp_dir):
      """
      Initiate the WAV offsetting algorithm.

      New output files are written to either :file:`wav[+,-]n/` or
      :file:`/tmp/mktoc.[random]/`

      :param files:  WAV files read to apply the sample shifting process to.
      :type files:   list

      :param use_tmp_dir:  :data:`True` indicates new WAV files are created in
                           :file:`/tmp`.
      :type use_tmp_dir:   bool
      """
      # initialize the progress bar class, set the maximum progress bar value
      self._pb = self._pb_class( bar_max=self._get_total_samp(files),
                                 *self._pb_args)
      # set the dir name generation function, and create out_file list
      if not use_tmp_dir: outdir = self._get_new_name
      else              : outdir = self._get_tmp_name
      out_files = map( outdir, files )

      # positive offset correction, insert silence in first track,
      # all other tracks insert end data of previous track
      if self._offset > 0:
         offsetter_fnct = self._insert_prv_end
         # create a list of 'previous' file names
         f2_list = [None] + files[:-1]
      # negative offset correction, append silence to end of last track,
      # all other tracks append start data of next track
      elif self._offset < 0:
         offsetter_fnct = self._append_nxt_start
         # create a list of 'next' file names
         f2_list = files[1:] + [None]

      map( offsetter_fnct, out_files, files, f2_list )
      # return a list of the new files names
      return out_files

   def _append_nxt_start(self, out_fn, fn, nxt_fn):
      """Negative offset correction algorithm for a single WAV file.
      Copies the current WAV file data and then appends start of the
      next files WAV data into a new WAV file. The basic steps are:
         1) perform a positive 'n' sample seek into input WAV file.
         2) copy data from location to start of new WAV file until EOF
            of input.
         3) Either,
            a) open 'nxt_fn' WAV file and finish writing the last 'n'
               samples.
            b) pad the new WAV file with n samples of NULL data.

      Parameters:
         out_fn   : String of output WAV file name.

         fn       : String of intput WAV file name.

         nxt_fn   : String of N+1 input WAV file name."""
      wav_out = wave.open(out_fn, 'w')
      wav_in = wave.open(fn)
      # setup the output parameters
      bytes_p_samp = wav_in.getsampwidth() * wav_in.getnchannels()
      offset_bytes = abs(self._offset) * bytes_p_samp
      wav_out.setparams( wav_in.getparams() )
      # seek ahead sample offset amount
      wav_in.setpos( abs(self._offset) )
      # copy all frame date from 1st file into new file
      while True:
         data = wav_in.readframes(self._COPY_SIZE)
         if len(data) == 0: break
         self._write_frames(wav_out, data, bytes_p_samp)
      wav_in.close()
      # finally copy the remaining data from the next track, or silence
      if nxt_fn:
         # copy offset frame date from next file into new file
         wav_in = wave.open(nxt_fn)
         data = wav_in.readframes( abs(self._offset) )
         assert len(data) == offset_bytes
         self._write_frames(wav_out, data, bytes_p_samp)
      else:
         # write silence to end of last track
         self._write_frames(wav_out, '\x00'*offset_bytes, bytes_p_samp)
      # print the progress bar
      wav_in.close()
      wav_out.close()

   def _get_new_name(self, f):
      """Generates a new name a location to write 'wav[+,-]n/' WAV
      files."""
      dir_,name = os.path.split(f)
      new_dir = os.path.join(dir_, 'wav%+d' % self._offset)
      if not os.path.exists(new_dir):
         os.mkdir( new_dir )
      return os.path.join( new_dir, name)

   def _get_total_samp(self, files):
      """Helper function to return the total sample count of a list of
      WAV files. Used to set the ProgressBar 'max' value.

      Parameter:
         files : List of WAV files to read."""
      return sum(itr.imap( lambda f: wave.open(f).getnframes(), files))

   def _get_tmp_name(self, f):
      """Generates a new name a location to write
      '/tmp/mktoc.[random]/' WAV files."""
      if not hasattr(self, '_tmp_dir'):
         self._tmp_dir = tempfile.mkdtemp( prefix=self._progName+'.' )
      return os.path.join( self._tmp_dir, os.path.basename(f) )

   def _insert_prv_end(self, out_fn, fn, prv_fn):
      """Positive offset correction algorithm for a single WAV file.
      Inserts the end of the previous files WAV data and then copies
      the current WAV files data into a new WAV file. The basic steps
      are:
         1) Either,
            a) perform a positive (EOF - 'n') sample seek into
               'prv_fn' WAV file.
            b) if no 'prv_fn' use NULL data
         2) copy n samples of data to start of new WAV file.
         3) Open 'fn' WAV file and copy the full WAV file - 'n'
            samples to the output WAV.

      Parameters:
         out_fn   : String of output WAV file name.

         fn       : String of intput WAV file name.

         prv_fn   : String of N-1 input WAV file name."""
      wav_out = wave.open(out_fn, 'w')
      wav_in = wave.open(fn)
      # setup the output parameters
      bytes_p_samp = wav_in.getsampwidth() * wav_in.getnchannels()
      offset_bytes = self._offset * bytes_p_samp
      wav_out.setparams( wav_in.getparams() )
      wav_in.close()
      # if previous file exists, insert end of stream to new file
      if prv_fn:
         wav_in = wave.open(prv_fn)
         pos = wav_in.getnframes() - self._offset   # seek position
         wav_in.setpos( pos ) # seek to EOF - offset
         data = wav_in.readframes( self._offset )
         assert len(data) == offset_bytes
         self._write_frames(wav_out, data ,bytes_p_samp)
         wav_in.close()
      else:    # insert silence if no previous file
         self._write_frames(wav_out ,'\x00'*offset_bytes, bytes_p_samp)
      # add original file data to output stream
      wav_in = wave.open( fn )
      samples = wav_in.getnframes() - self._offset
      while samples:
         data = wav_in.readframes( min(samples,self._COPY_SIZE) )
         samples -= len(data) / bytes_p_samp
         self._write_frames(wav_out, data, bytes_p_samp)
      wav_in.close()
      wav_out.close()

   def _write_frames(self, fh, data,bps):
      """Wrapper for writing data wav files. A secondary side effect
      is that each call udpates the progress bar."""
      fh.writeframes(data)
      self._pb += len(data) / bps      # update progress bar
      sys.stderr.write(str(self._pb))  # print the progress bar




#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   mktoc.cmdline
   ~~~~~~~~~~~~~

   Command-line interface for Mktoc.
"""

import codecs
import chardet.universaldetector
import logging
import os
import re
import sys
import textwrap
import traceback
from optparse import OptionParser

from .base import *
from .parser import *


# WAV file reading command-line switch
#  - allow 'file not found' errors when reading WAV files
_OPT_ALLOW_WAV_FNF   = '-a'
# Offset correction command-line switch
# - enable output of offset corrected WAV files
_OPT_OFFSET_CORRECT  = '-c'
# Output CUE file name
_OPT_CUE_FILE        = '-f'
# Temp WAV Files
# - write offset corrected WAV files to /tmp dir
_OPT_TEMP_WAV        = '-t'
# WAV File List
# - create a TOC file using a list of WAV files
_OPT_WAV_LIST        = '-w'
# Mulitsession TOC
# - required when writting a multi-session TOC
_OPT_MULTI_SESSION  = '-m'
# - disable multi-session features, don't prompt
_OPT_IGNORE_MULTI_SESSION = '-z'


class CommandLine(object):
   """
   Command line runner class for executing Mktoc.

   Interprets all program arguments and creates a CueParser object to
   generate a final TOC file.
   """
   def run(self,argv=sys.argv[1:]):
      """Execution entry point."""
      try:
         self._run(argv)
      except TooManyFilesMatchError as e:
         self._error_msg_multi_files(e)
      except FileNotFoundError as e:
         self._error_msg_no_file(e)
      except MkTocError, e:
         self._error_msg(e)

   def _run(self,argv):
      # parse all command line arguments, exit if there is any error
      opt,args = self._parse_args(argv)
      # setup logging
      if opt.debug: logging.basicConfig(level=logging.DEBUG)
      # check if using WAV list or CUE file
      if opt.wav_files is None:
         # open CUE file
         if opt.cue_file:
            fh_in = self._open_file( opt.cue_file )
            # set the working dir of the input file
            cue_dir = os.path.dirname( fh_in.name ) or os.curdir
         else:
            cue_dir = os.curdir
            fh_in = sys.stdin
         # create CUE file parser
         assert(cue_dir)
         p = CueParser( cue_dir, opt.find_wav)
         cd_obj = p.parse( fh_in)
         fh_in.close()
      else:
         wav_dir = os.path.dirname( opt.wav_files[0] ) or os.curdir
         # create WAV list parser
         assert( wav_dir)
         p = WavParser( wav_dir, opt.find_wav)
         cd_obj = p.parse( opt.wav_files)
      # warn user when TOC is multi-session
      self._check_multisession_opt( cd_obj, opt)
      if opt.wav_offset:
         cd_obj.modWavOffset( opt.wav_offset, opt.write_tmp )
      toc = cd_obj.getToc()
      # open TOC file
      if opt.toc_file:
         fh_out = self._open_file( opt.toc_file,'wb')
      else:
         fh_out = sys.stdout
      fh_out.write( self._banner_msg())
      for l in toc:
         fh_out.write("%s\n" % l)
      fh_out.close()

      if cd_obj.disc.is_multisession:
         # print multi-session instructions; data session size is calulated by
         # frame length minus 2 frames. I'm not actually sure why 2 frames must
         # be subtracked, but it was verify to be correct. If your system/drive
         # behaves differntly, please file a bug report.
         print >> sys.stderr, textwrap.dedent("""
         #########################################################
         # Multi-Session TOC Mode
         #########################################################

         1. Burn TOC file w/ cdrdao '--multi' option
         2. Finalize disc with dummy data session command:

            cdrecord --tsize=%ds /dev/zero

         #########################################################
         """ % (cd_obj.last_index.len_.frames-2))    # see note for '-2'

   def _open_file(self,name,mode='rb',encoding=None):
      """Wrapper for opening files. Ensures correct encoding is selected."""
      try:
         if encoding is None:
            # detect file character encoding
            with open(name,mode) as fh:
               d = chardet.universaldetector.UniversalDetector()
               for line in fh.readlines():
                  d.feed(line)
               d.close()
               encoding = d.result['encoding']
         return codecs.open(name, mode, encoding=encoding)
      except:
         print >> sys.stderr, sys.exc_value
         exit(-1)

   def _check_multisession_opt(self, cd, opt):
      """Check multi-session run-time options match track info."""
      if not cd.disc.is_multisession:
         return
      if opt.no_multisession:
         # disable multi-session
         cd.disc.is_multisession = False
      elif not opt.multisession:
         # multisesssion option must be set to prevent usage error
         print >> sys.stderr, textwrap.dedent("""
            WARNING! - Detected multi-session track info.

            For safety, '%s' option must be specified when creating a TOC
            for a multi-session disc.

            If you want to ignore this check, and disable multi-session
            features, use the '%s' argument.""" %
               (_OPT_MULTI_SESSION,_OPT_IGNORE_MULTI_SESSION))
         sys.exit(-1)

   def _banner_msg(self):
      """Returns a TOC comment header that is placed at the top of the
      TOC file."""
      return "// Generated by %s %s\n" % (progName, VERSION) + \
         "// %s, %s\n" % (__copyright__, __author__) + \
         "// Report bugs to <%s>\n" % __email__

   def _parse_args(self,argv):
      """Use OptionParser object to handle all input arguments and
      return opt structure and args list as a tuple. All argument
      error checking is performed in this function."""
      usage = '[OPTIONS] [[-f] CUE_FILE|-w WAV_FILES] [[-o] TOC_FILE]'
      parser = OptionParser( usage='%prog '+usage, version='%prog '+VERSION,
                             conflict_handler='resolve')
      parser.add_option('--help', action='callback',
            callback=self._parse_full_help,
            help='show detailed usage instructions and exit' )
      parser.add_option( _OPT_ALLOW_WAV_FNF, '--allow-missing-wav',
            dest='find_wav', action="store_false", default=True,
            help='do not abort when WAV file(s) are missing, (experts only)')
      parser.add_option( _OPT_OFFSET_CORRECT, '--offset-correction',
            dest='wav_offset', type='int',
            help='correct reader/writer offset by creating WAV file(s) '
                 'shifted by WAV_OFFSET samples (original data is '
                 'not modified)' )
      parser.add_option('-d', '--debug', dest='debug', action="store_true",
            default=False, help='enable debugging statements' )
      parser.add_option( _OPT_CUE_FILE, '--file', dest='cue_file',
            help='specify the input CUE file to read')
      parser.add_option( _OPT_MULTI_SESSION, '--multi', dest='multisession',
            action='store_true', default=False,
            help='for safety, this option must be set when creating a '
                 'mulit-session TOC file' )
      parser.add_option('-o', '--output', dest='toc_file',
            help='specify the output TOC file to write')
      parser.add_option( _OPT_TEMP_WAV, '--use-temp', dest='write_tmp',
            action='store_true', default=False,
            help='write offset corrected WAV files to /tmp directory' )
      parser.add_option( _OPT_WAV_LIST, '--wave', dest='wav_files',
            action='callback', callback=self._parse_wav,
            help='write a TOC file using list of WAV files' )
      parser.add_option( _OPT_IGNORE_MULTI_SESSION, '--no-multi',
            dest='no_multisession', action='store_true', default=False,
            help='disable multi-session support; program assumes TOC will be '
                 'written in single-session mode' )
      # execute parsing step
      opt,args = parser.parse_args(argv)

      # test "WAV file not found" and "offset correction" argument combination
      if opt.wav_offset and not opt.find_wav:
         parser.error("Can not combine '%s' and '%s' options!" % \
                        (_OPT_ALLOW_WAV_FNF,_OPT_OFFSET_CORRECT) )
      # test "offset correction" and "temp WAV" argument combination
      if opt.write_tmp and not opt.wav_offset:
         parser.error("Can not use '%s' without '%s' option!" % \
                        (_OPT_TEMP_WAV, _OPT_OFFSET_CORRECT) )
      # test "CUE File" and "-w" argument combination
      if opt.cue_file is not None and opt.wav_files is not None:
         parser.error("Can not combine '%s' and '%s' options!" % \
                        (_OPT_CUE_FILE, _OPT_WAV_LIST) )
      # test "--multi" and "--no-multi" argument combination
      if opt.multisession and opt.no_multisession:
         parser.error("Can not combine '%s' and '%s' options!" % \
                        (_OPT_MULTI_SESSION, _OPT_IGNORE_MULTI_SESSION) )
      # The '-w' option is used to create a TOC file using a list of WAV files.
      # The default mode is to convert a CUE file. The 'if' checks for the
      # default mode.
      if opt.wav_files is None:
         # the '-w' flag is not set. 1st argument names the input CUE file, 2nd
         # argument names the output TOC file.
         if len(args)>2 or \
               (len(args)>=1 and opt.cue_file) or \
               (len(args)>=2 and opt.toc_file):
            parser.error("Ambiguous file arguments!")
         else:
            # set file names if 'args' list is not empty
            if len(args)>=1: opt.cue_file = args[0]
            if len(args)>=2: opt.toc_file = args[1]
      else:
         # the '-w' flag was set. Only 1 argument is allowed to name the output
         # TOC file.
         if len(args)>1:
            parser.error("Ambiguous file arguments!")
         else:
            # set file names if 'args' list is not empty
            if len(args)>=1: opt.toc_file = args[0]
      return opt,args

   def _parse_full_help(self, option, opt_str, value, parser):
      """Callback to print a detailed help output page."""
      import mktoc
      parser.exit( msg=mktoc.__doc__)

   def _parse_wav(self, option, opt_str, value, parser):
      """OptionParser callback function to correctly handle '-w' WAV
      file input arguments. All trailing WAV files will be placed in
      the parser.values.wav_files list. The operation will stop at
      either the first non-WAV file argument or the end of the
      argument list."""
      slice_idx = 0
      for arg in parser.rargs:
         if re.search( r'\.wav$', arg, re.IGNORECASE):
            slice_idx += 1
         else: break

      # if one or more WAV files were found, move them to the parser wav_file
      # list
      if slice_idx:
         parser.values.wav_files = parser.rargs[0:slice_idx]
         del parser.rargs[0:slice_idx]
      else:
         parser.error( '%s option requires one or more WAV file arguments' \
                        % opt_str )

   def _error_msg(self, e):
      """Print a default error message to the user."""
      print >> sys.stderr, textwrap.dedent("""
      ERROR! -- An unrecoverable error has occurred.

      If you believe the CUE file is correct, please send the input file to
      <%s>, along with the error message below.

      ---> %s
      """ % (__email__,e))

   def _error_msg_multi_files(self, e):
      """Print error when duplicate WAV files are found."""
      print >> sys.stderr, textwrap.dedent( """
      ERROR! -- Could not resolve WAV file:
         '%s'\n""" % (e.src_file,))

      print >> sys.stderr, "   Conflicting matches are:"
      for f in e.found_files:
         print >> sys.stderr, '      ' + f

      print >> sys.stderr, textwrap.dedent( """
      Cdrdao can not correctly write pregaps in TOC files without explicit
      file lengths. If you know what you are doing, you can disable this
      check with the '%s' option.""" % (_OPT_ALLOW_WAV_FNF,))

   def _error_msg_no_file(self, e):
      """Print a missing WAV file error message to the user."""
      print >> sys.stderr, textwrap.dedent( """
      ERROR! -- Could not find the WAV file:
         '%s'

      Cdrdao can not correctly write pregaps in TOC files without explicit
      file lengths. If you know what you are doing, you can disable this
      check with the '%s' option.""" % (e,_OPT_ALLOW_WAV_FNF,))


def main():
   """
   Primary entry point for the mktoc command line application.

   Creates a :class:`CommandLine` object and catches any exceptions. Returns 0
   to indicate success, or any other value for failure.
   """
   global progName
   progName = os.path.basename(sys.argv[0])
   try:
      CommandLine().run()
   except EmptyCueData: pass     # ignore NULL data input (Ctrl-C)
   except Exception:
      traceback.print_exc()
   except: pass      # ignore base exceptions (exit,key-int)
   else: return 0    # no exception, exit success
   return 1          # exit with failure



#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   mktoc.fsm
   ~~~~~~~~~

   A finite state machine implementation.
"""

import operator as op


class NullStateException(Exception):
   """
   Exception thrown when :class:`StateMachine` has no state defined to handle
   the current input data.
   """


class StateMachine(object):
   """
   Base class for building a finite state machine.

   The state machine is controled by two externally provided data structures.

   .. attribute:: regex_obj

      A compiled :mod:`regex <re>` pattern instance that contains one or more
      named groups (i.e. ``(?P<name>...)``) sub-patterns.

   .. attribute:: match_handlers

      A :class:`dict` that maps a handler function to a group name returned by
      a successful match in the :data:`regex_obj` instance.

   .. warning::

      The :meth:`change_state` method must be run to initialize the state
      machine before the :class:`StateMachine` instance is called.

   .. Document private members
   .. automethod:: __call__
   """
   __regex_obj = None
   __match_handlers = None

   def __call__(self,lines):
      """
      For each line of text in :data:`lines`, a regex match is performed
      and a state handler function is called to finalize processing of the
      string.

      :param lines: Input data consumed by state machine
      :type  lines: list
      """
      for l in lines:
         match = self.__regex_obj.match(l)
         if match:
            match_name = match.lastgroup
            match_groups = [x for x in match.groups() if x]
            self.__match_handlers[match_name]( match_name, *match_groups)
         else:
            raise NullStateException(repr(l))

   def change_state(self,regex_obj=None,match_handlers=None):
      """
      Modifies the control flow of the state machine.

      Replaces the internal :data:`regex_obj` or :data:`match_handlers`.
      Using keywords, one paramter can be changed without effecting the other.

      :param regex_obj: Compiled regular expression object
      :type  regex_obj: :ref:`regex <re-objects>`

      :param match_handlers: handler functions map, keyed with
                             :data:`regex_obj` group name.
      :type  match_handlers: :class:`dict` of :func:`callable`\'s
      """
      if regex_obj:
         self.__regex_obj = regex_obj
      if match_handlers:
         self.__match_handlers = match_handlers




#!/usr/bin/env python

#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   Distools setup file for the mktoc application.
"""

import os
from setuptools import setup
from textwrap import dedent

import mktoc
from mktoc.base import *

setup(
   name='mktoc',
   version=VERSION,
   description= 'Simple command line tool generates TOC files '
                'for audio CD burning with cdrdao.',
   long_description=dedent(mktoc.__doc__),
   author=__author__,
   author_email=__email__,
   url='http://packages.python.org/mktoc/',
   download_url=(
      'https://github.com/cmcginty/mktoc/raw/master/dist/mktoc-%s.tar.gz'
      % (VERSION,)),
   packages=['mktoc'],
   entry_points = {
      'console_scripts': ['mktoc = mktoc.cmdline:main',],
   },
   keywords = 'cdrdao cue toc cd-writing audio-cd mktoc mkcue cue2toc '
              'eac exactaudiocopy pregap',
   classifiers=[
      'Development Status :: 5 - Production/Stable',
      'Environment :: Console',
      'Intended Audience :: End Users/Desktop',
      'License :: OSI Approved :: GNU General Public License (GPL)',
      'Operating System :: OS Independent',
      'Operating System :: POSIX :: Linux',
      'Programming Language :: Python',
      'Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Writing',
      'Topic :: Multimedia :: Sound/Audio :: Conversion',
   ]
)




#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   mktoc.parser
   ~~~~~~~~~~~~

   This module provides object(s) to parse text files describing the layout of
   an audio CD. After the parse step is complete, it is possible to access the
   data or convert into any other output format.

   The following are a list of the classes provided in this module:

   * :class:`ParseData`
   * :class:`CueParser`
   * :class:`WavParser`
"""

from __future__ import absolute_import

from itertools import *
import logging
import operator as op
import os
import re

from .base import *
from . import disc
from . import wav
from . import fsm
from . import progress_bar

__all__ = ['CueParser','WavParser']

log = logging.getLogger('mktoc.parser')


class ParseData(object):
   """
   Stores parsed CD-ROM data and provides methods for modifcation and access.

   Automatically generated by invoking the :meth:`parse` method defined in one
   of the :class:`_Parser` classes.
   """
   def __init__(self, disc, tracks, files):
      """
      Initialize data structures.

      :param disc: CD info data object
      :type  disc: :class:`~mktoc.disc.Disc`

      :param tracks: a lost of objects with track info and indexes for each
                     portion of the track.
      :type  tracks: :func:`list` of class:`disc.Track`

      :param files:  in-order list of WAV files associated with 'tracks'
      :type  files:  :func:`list` of file name str\s
      """
      if len(tracks) == 0:
         raise ParseError()
      self.disc    = disc   # disc object that stores global disc info.
      self._tracks = tracks # track object that stores track info.
      self._files  = files  # in-order list of WAV files that apply to the CD
                            # audio.

   @property
   def last_index(self):
      """Reference to last index of last track."""
      assert self.disc.is_multisession
      return self._tracks[-1].indexes[-1]

   def getToc(self):
      """
      Access method to return a text stream of the CUE data in TOC format.
      """
      toc = []
      toc.extend( str(self.disc).split('\n') )
      for trk in self._tracks:
         toc.extend( str(trk).split('\n') )
      # expand tabs to 4 spaces, strip trailing white space on each line
      toc = [line.expandtabs(4).rstrip() for line in toc]
      return toc

   def modWavOffset(self,samples,tmp=False):
      """
      Optional method to correct the audio WAV data by shifting the samples by
      a positive or negative offset.

      This can be used to compensate for a write offset in a CD/DVD burner. If
      the `tmp` variable is :data:`True`, all new WAV files will be created in
      the :file;`/tmp` directory.

      :param samples:   Number of samples to shift the audio data by. This
                        value can be negative or positive.
      :type  samples:   int

      :param tmp:    :data:`True` or :data:`False`; when :data:`True` any
                     new WAV files will be created in :file:`/tmp`.
      :type tmp:     bool
      """
      # create WavOffset object, initialize sample offset and progress output
      wo = wav.WavOffsetWriter( samples, progress_bar.ProgressBar,
                                   ('processing WAV files:',))
      new_files = wo( self._files, tmp )

      # change all index file names to newly generated files
      file_map = dict( zip(self._files,new_files) )
      indexes = imap(op.attrgetter('indexes'), self._tracks);
      for idx in chain(*indexes):
         if idx.file_: # data tracks do not have valid files
            log.debug( "updating index file '%s'", idx.file_ )
            idx.file_ = file_map[idx.file_]


class _FileLookup(object):
   """
   Return the path to a valid WAV file in the files system using the input
   :param:`file_` value.

   If the WAV file can not be found and :param:`_find_wav` is :data:`True`,
   then an exception is raised.
   """
   # Dictionary to map input WAV files to actual files on the system. The map
   # is for use in cases where the defined file name does not exactly match the
   # file system WAV name.
   _file_map         = None

   # True or flase, when True the WAV file must be found in the FS or an
   # exception is raised.
   _find_wav         = None

   # WavFileCache object that can quickly find WAV files in the local file
   # system.
   _wav_file_cache   = None

   def __init__(self, dir_, find_wav):
      """
      :param dir_:      Path location of the working directory
      :type  dir_:      string

      :param find_wav:  :data:`True`/:data:`False1, :data:`True` causes
                        exceptions to be raised if a WAV file can not be found
                        in the FS.
      :type  find_wav:  bool

      .. Document private members
      .. automethod:: __call__
      """
      # init class options
      self._dir            = dir_
      self._find_wav       = find_wav
      self._file_map       = {}
      assert(dir_)
      self._wav_file_cache = wav.WavFileCache(dir_)

   def __call__(self,file_):
      """
      :param file:   Audio file name parsed from the CUE text.
      :type  file:   string
      """
      if file_ in self._file_map:
         return self._file_map[file_]
      else:
         try:  # attempt to find the WAV file
            file_on_disk = self._wav_file_cache(file_)
         except FileNotFoundError:
            # raise only if '_find_wav' option is True
            if self._find_wav: raise
            else: file_on_disk = file_
         self._file_map[file_] = file_on_disk
         return file_on_disk


class _CueStateMachine(fsm.StateMachine):
   """
   State machine logic for parsing CUE commands in a CUE file.
   """
   #: Regex match pattern for CUE command syntax
   CUE_CMDS = re.compile( r"""
      (?P<catalog>
         ^CATALOG             # CATALOG
            \s+(\d{13})$) |   # value

      (?P<flags>
         ^FLAGS               # FLAG
         \s+(.*)$) |          # one or more flags

      (?P<file>
         ^FILE                # FILE
         \s+"(.*)"            # 'file name' in quotes
         \s+WAVE$) |          # WAVE

      (?P<index>
         ^INDEX                        # INDEX
         \s+(\d+)                      # 'index number'
         \s+(\d{2}:\d{2}:\d{2})$) |    # 'index time'

      (?P<isrc>
         ^ISRC                # ISRC
         \s+(.*)$) |          # value

      (?P<performer>
         ^PERFORMER           # PERFORMER
         \s+"(.*)"$) |        # quoted string

      (?P<pregap>
         ^PREGAP              # PREGAP
         \s+(.*)$) |          # value

      (?P<title>
         ^TITLE               # TITLE
         \s+"(.*)"$) |        # quoted string

      (?P<track>
         ^TRACK                  # TRACK
         \s+(\d+)                # track 'number'
         \s+(AUDIO|MODE.*)$) |   # AUDIO or MODEx/xxxx

      (?P<rem>
         ^REM                 # REM
         \s*(\w*)             # sub-keyword
         \s*(.*))             # remaining text
      """, re.VERBOSE)

   def __init__(self, file_lookup, dir_):
      """
      :param file_lookup:  Callable instance for quickly correlating files in the
                           local file system from file names in CUE commands.
      :type  file_lookup:  :class:`_FileLookup`

      :param dir_:   Path location of the working directory.
      :type  dir_:   str

      .. Document private members
      .. automethod:: __call__
      """
      # callback mapping for 'DISC' state commands
      self.disc_handlers = {
         'catalog'      : self.cmd_field_disc,
         'file'         : self.cmd_file,
         'performer'    : self.cmd_field_disc,
         'rem'          : self.cmd_rem,
         'title'        : self.cmd_field_disc,
         }
      # callback mapping for 'FILE' state commands
      self.file_handlers = {
         'file'         : self.cmd_file,
         'index'        : self.cmd_index,
         'track'        : self.cmd_track,
         }
      # callback mapping for 'TRACK' state commands
      self.track_handlers = {
         'file'         : self.cmd_file,
         'flags'        : self.cmd_flags,
         'index'        : self.cmd_index,
         'isrc'         : self.cmd_field_trk,
         'performer'    : self.cmd_field_trk,
         'pregap'       : self.cmd_field_trk,
         'rem'          : self.cmd_noop,
         'title'        : self.cmd_field_trk,
         'title'        : self.cmd_field_trk,
         'track'        : self.cmd_track,
         }
      # instance variables for managing parsing logic
      self.disc   = disc.Disc()
      self.tracks = []
      self.track  = None
      self.files  = []
      self.file_  = None
      self.file_lookup = file_lookup
      self.dir_   = dir_
      # initialize beginning state
      self.change_state( self.CUE_CMDS, self.disc_handlers )

   def __call__(self,*a,**kw):
      """
      Extends the super class method by catching 'KeyErrors' caused by
      unexpected or unmatched patterns.
      """
      try:
         super(_CueStateMachine,self).__call__(*a,**kw)
      except (fsm.NullStateException,) as e:
         raise ParseError( 'Unknown/invalid command: ' + str(e) )
      return ParseData(self.disc, self.tracks, self.files)

   def cmd_noop( self, match_name, cmd, *args ):
      """Ignored commands"""

   def cmd_rem( self, match_name, cmd, field, value):
      """
      Store a REM field in the disc data, unhandled fields will be silently
      ignored.
      """
      self.disc.set_field( field, value)

   def cmd_field_disc( self, match_name, cmd, value):
      """Store a command field in the disc data."""
      self.disc.set_field( match_name, value)

   def cmd_field_trk( self, match_name, cmd, value):
      """Store a command field in the track data."""
      self.track.set_field( match_name, value)

   def cmd_file( self, match_name, cmd, file_):
      """Process a new data file name. Changes state to 'FILE'."""
      self.file_ = self.file_lookup(file_)
      self.files.append( self.file_ )
      self.change_state( match_handlers=self.file_handlers )  # next state

   def cmd_track( self, match_name, cmd, trk_num, trk_type):
      """Create a new :class:`~mktoc.disc.Track` instance.
      Change state to 'TRACK'.
      """
      self.track = disc.Track(int(trk_num), trk_type != 'AUDIO')
      self.tracks.append( self.track )
      if trk_type != 'AUDIO':
         self.disc.is_multisession = True    # disc is multi-session
      self.change_state( match_handlers=self.track_handlers ) # next state

   def cmd_index( self, match_name, cmd, idx_num, time):
      """
      Create a new :class:`~mktoc.disc.TrackIndex` instance.

      Additional processing steps are performed here to modify the state of
      previous data structures.
      """
      if not self.track.is_data:
         idx = disc.TrackIndex( idx_num, time, self.file_)
      else:
         # if data track, the length is not defined in the CUE and must be
         # sourced from another method to create a 100% accurate TOC
         size = self.data_trk_size(self.track.num)
         if not size:
            # TODO: the user should be allowed to override this error in the
            # future, and/or have feature to manually set the track size
            raise ParseError('size of DATA track can not be determined')
         idx = disc.TrackIndex( idx_num, time, None, size)
         idx.cmd = disc.TrackIndex.DATA

      self.track.indexes.append( idx )

      # set local var defaults
      prev_idx = None
      if len(self.track.indexes) >= 2:
         prev_idx = self.track.indexes[-2] # [-1] is current index
      prev_trk = None
      if len(self.tracks) >= 2:
         prev_trk = self.tracks[-2] # [-1] is current track

      # Add 'START' command after pregap audio file
      #
      # if 'prev_idx' is a track pregap (num == 0) and the file for 'current
      # index' is not the same, then designate the 'prev_idx' as a pregap audio
      # only. The result is to place a TOC 'START' command between the
      # 'prev_idx' and the 'current index' in the TOC file.
      if prev_idx and prev_idx.num == 0 and idx.file_ != prev_idx.file_:
         prev_idx.cmd = disc.TrackIndex.PREAUDIO

      # When a single WAV file is used for multiple internal track indexes:
      if prev_idx and self.file_ == prev_idx.file_:
         if prev_idx.num == 0:
            # Designate the 'true' start index of a track when the track data
            # file contains pregap data. This is done with the TOC command
            # 'START'
            #
            # details:  if the current index is the pregap data (0), then the
            #           pregap must be set by changing the 'next' index cmd to
            #           'START', and the length of the pregap must be set.
            idx.cmd = disc.TrackIndex.START
            idx.len_ = idx.time - prev_idx.time
            del idx.time # remove for safety, do not use
         else:
            # Else not a pregap, change the TOC command for a new track to
            # 'INDEX' when a single logical 'track' has multiple index values
            # (by default, the TOC command is AUDIOFILE when a track has a
            # single index).
            #
            # details:  the outside 'if' guarantee that the current and next
            #           index use the same file. Also, since it is not a
            #           pregap the TOC format must use 'INDEX' keyword
            #           instead of AUDIOFILE. No other calculations are
            #           needed because INDEX is specified by file offset.
            idx.cmd = disc.TrackIndex.INDEX
            del idx.len_ # remove for safety, do not use

      # Set the LENGTH argument on a track fle that must stop before EOF
      #
      # On the current index, which is the first index of track 2 or
      # greater,
      if prev_trk and len(self.track.indexes) == 1:
         for prev_idx in prev_trk.indexes:
            # if TOC command for previous track index is AUDIOFILE, and if prev
            # track uses the same file, then prev INDEX must end before the
            # current track INDEX starts.
            if (prev_idx.cmd == disc.TrackIndex.AUDIO
                  and prev_idx.file_ == self.file_):
               prev_idx.len_ = idx.time - prev_idx.time

   def cmd_flags( self, match_name, cmd, flags):
      """Set the state of flag fields in a :class:`disc.Track` instance."""
      for f in filter(lambda x: x in ['DCP','4CH','PRE'],flags.split()):
         if f == '4CH': f = 'four_ch'     # change '4CH' flag name
         self.track.set_field(f,True)

   def data_trk_size(self, trk_idx):
      """
      Use an ExactAudioCopy log file to determine the length of the track at
      the specified index.

      :param trk_idx: Track index of data
      :type  trk_idx: int
      """
      import codecs
      import chardet.universaldetector
      size = None
      files = os.listdir(self.dir_)
      logs = [f for f in files if os.path.splitext(f)[1] == '.log']
      logs.sort()
      for f in logs:
         # detect file character encoding
         with open(os.path.join(self.dir_,f),'rb') as fh:
            d = chardet.universaldetector.UniversalDetector()
            for line in fh.readlines():
               d.feed(line)
            d.close()
            encoding = d.result['encoding']
         with codecs.open( os.path.join(self.dir_,f),
                           'rb', encoding=encoding) as fh:
            lines = fh.readlines()
         regex = re.compile(r'^\s+%d\s+\|.+\|\s+(.+)\s+\|.+\|.+$' % (trk_idx,))
         matches = filter(None,map(regex.match,lines))
         if matches:
            # convert first match from '1:11.11' to '1:11:11'
            size = matches[0].group(1).replace('.',':')
            break

      return size


class CueParser(object):
   """
   An audio CUE sheet text file parsing class.

   By matching the known format of a CUE file, the relevant text information is
   extracted and converted to a binary representation. The binary
   representation is created by using combination of Disc, Track, and
   TrackIndex objects. With the data, the CUE file can be re-created or
   converted into a new format.
   """
   def __init__(self, dir_=os.curdir, find_wav=True):
      """
      :param dir_:  Path location of the CUE file's directory.
      :type  dir_:  str

      :param find_wav:  :data:`True`/:data:`False`, :data:`True` causes
                        exceptions to be raised if a WAV file can not be found
                        in the FS.
      :type  find_wav:  bool
      """
      self.dir_ = dir_
      self.file_lookup = _FileLookup(dir_,find_wav)

   def parse(self, fh):
      """
      Parses CUE file text data.

      :param fh:  An open file handle used to read the CUE text data
      :type fh:   :data:`file`

      :returns: :class:`ParseData` instance that mirrors the CUE data.
      """
      # parse disc into memory, ignore comments
      cue = [line.strip() for line in fh]
      if not len(cue):
         raise EmptyCueData
      # begin state machine in 'Init' state
      csm = _CueStateMachine(self.file_lookup, self.dir_)
      return csm( cue )


class WavParser(object):
   """
   A simple parser object that uses a list of WAV files to create a CD TOC.

   The class assumes that each WAV file is an individual track, in ascending
   order.
   """
   def __init__(self, dir_=os.curdir, find_wav=True):
      """
      :param dir_:  Path location of the CUE file's directory.
      :type  dir_:  str

      :param find_wav:  :data:`True`/:data:`False`, :data:`True` causes
                        exceptions to be raised if a WAV file can not be found
                        in the FS.
      :type find_wav: bool
      """
      # init class options
      self.file_lookup = _FileLookup(dir_,find_wav)

   def parse( self, wav_files):
      """
      Parses a list of WAV files.

      :param wav_files: WAV files to add to the TOC
      :type  wav_files: list

      :returns: :class:`ParseData` instance that mirrors the WAV data.
      """
      files = map(self.file_lookup, wav_files)
      # return a new Track object with a single Index using 'file_'
      def mk_track((idx,file_)):
         # create a new track for the WAV file
         trk = disc.Track(idx+1)
         # add the WAV file to the first index in the track
         trk.indexes.append( disc.TrackIndex(1,0,file_) )
         return trk
      # return a new ParseData object with empy Disc and complete Track list
      return ParseData( disc.Disc(),
                        map( mk_track, enumerate(files)), files )




#  Copyright (c) 2011, Patrick C. McGinty
#
#  This program is free software: you can redistribute it and/or modify it
#  under the terms of the Simplified BSD License.
#
#  See LICENSE text for more details.
"""
   mktoc.progress_bar
   ~~~~~~~~~~~~~~~~~~

   Module for mktoc that prints a progress indication.

   The default usage is to prompt the user when an operation is running that
   the user must wait for. The following object classes are:
"""

import time

from mktoc.base import *

__all__ = ['ProgressBar']


##############################################################################
class ProgressBar( object ):
   """
   Creates a progress bar string to be printed by the calling function.
   """

   #: The maximum input input value expected by the progress bar. This value
   #: should be set before trying to print the progress bar. All percentage
   #: calculations are based from this value. It is OK to update this value as
   #: many times as needed, however it might confuse the user.
   bar_max = None

   # String to contain a message printed alongside the progress bar.
   _notice_text = None

   # The total integer count of the 'progress'. This value is modified by the
   # overloaded '+=' operator. This value can never go above 'bar_max'.
   _size = None

   def __init__(self, notice_txt, bar_max=0):
      """
      Initialize object defaults.

      :param notice_txt:   Message printed alongside the progress bar.
      :type notice_txt:    str

      :param bar_max:   Maximum size of the progress bar class.
      :type bar_max:    int
      """
      self._notice_txt = notice_txt
      self._size = 0
      self.bar_max = bar_max

   def __iadd__(self, other):
      """+= operator that increments the current state of the progress bar. The
      input value can be of any range, but the progress bar value will be fixed
      at 'bar_max'."""
      self._size += min(other, self.bar_max - self._size)
      return self

   def __str__(self):
      """Returns a progress bar string."""
      if not self.bar_max:
         raise Exception("You must initialize ProgressBar.bar_max first")
      if not hasattr(self,'_start_time'):
         self._start_time = time.time()
         time_dif = 0
      else:
         time_dif = time.time() - self._start_time    # compute time from start
      percent = float(self._size) / self.bar_max * 100
      if time_dif:
         rate = self._size / time_dif      # calculate sample/sec
         # estimate time left
         remain_time = (self.bar_max - self._size) / rate
         remain_str = '\tETA [%d:%02d]' % divmod(remain_time,60)
      else:
         remain_str = '\tETA [?:??]'
      return '%s %3d%% %s\r' % (self._notice_txt, percent, remain_str)

